#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "../GPU_Lib.h"
#include "../GPU_ExpEval.h"
#include "../SpeedTest/Utils.h" // p_LoadImage, Save_RGBLinearImageStruct

int main()
{
	if(!Init_OpenGL()) {
		fprintf(stderr, "error: OpenGL initialization failed\n");
		return -1;
	}
	// init openGL

	ImageStruct *p_src0, *p_src1;
	if((p_src0 = p_LoadImage("_lena_256.jpg")) == NULL ||
	   (p_src1 = p_LoadImage("_edgefinder_a.tga")) == NULL)
		return -1;
	// load input images

	ImageStruct *p_result;
	if((p_result = p_Alloc_Image(p_src0->XSize, p_src0->YSize, p_src0->PixelType)) == NULL) {
		fprintf(stderr, "error: error creating ImageStruct for result\n");
		return -1;
	}
	// allocate dest image

	float p_set_matrix[] = {
		1, 0, 0, 0,
		1, 0, 0, 0,
		1, 0, 0, 0,
		1, 0, 0, 0
	};
	// matrix for setting all color planes equal

	float p_sum_matrix[] = {
		1, 1, 1, 1,
		1, 1, 1, 1,
		1, 1, 1, 1,
		0, 0, 0, 0
	};
	// matrix for summing up color planes

	char *p_s_expression = "invert(hitmiss(thresh(grey(%0), .5) * #1, %2) * #3)";
	void *p_params[] = {p_src0, &p_set_matrix, p_src1, &p_sum_matrix};
	// expression and corresponding params vector

	printf("execute equation:\n> %s\n", p_s_expression);
	// prints expression to execute

	char *p_s_error;
	unsigned int n_string_pos;
	TExpressionHandle *p_expr = p_ParseExpr(p_s_expression, p_params, &p_s_error, &n_string_pos);
	// parse expression and ...

	if(p_s_error != NULL) {
		printf("error: %s\n> %s\n  ", p_s_error, p_s_expression);
		for(unsigned int i = 0; i < n_string_pos; ++ i)
			printf(" ");
		printf("^\n");
		// ... show where is error
	} else if(p_expr != NULL) {
		if(ImageExpression(p_result, p_expr, p_params)) {
			if(!Save_RGBLinearImageStruct("edges_lena.bmp", p_result))
				fprintf(stderr, "error: error writing output image\n");
			if(b_OpenGLError())
				fprintf(stderr, "error: OpenGL errors occured\n");
			else
				printf("done\n");
		} else
			fprintf(stderr, "error: expression execution failed\n");
		// ... or execute and save image
	}

	if(p_expr != NULL)
		FreeHandle(p_expr);
	// return expression handle

	p_src0->Optional.Regular.Free(p_src0);
	p_src1->Optional.Regular.Free(p_src1);
	p_result->Optional.Regular.Free(p_result);
	// defete images (in case GPU_Lib allocated them,
	// don't use your own code to free them)

	getchar();

	if(!Shutdown_OpenGL()) {
		fprintf(stderr, "error: OpenGL shutdown failed");
		return -1;
	}

	return 0;
}
