/*
								+----------------------------------+
								|                                  |
								| *** Image transfer functions *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|         ImageTransfer.h          |
								|                                  |
								+----------------------------------+
*/

#ifndef __IMAGE_TRANSFER_INCLUDED
#define __IMAGE_TRANSFER_INCLUDED

/**
 *	unsigned int n_PixelSize(int n_pixel_type)
 *		- returns size of DigiLib pixel type n_pixel_type
 */
unsigned int n_PixelSize(int n_pixel_type);

/**
 *	unsigned int n_GL_PixelSize(GLenum n_internal_format)
 *		- returns size of OpenGL pixel type n_internal_format
 */
unsigned int n_GL_PixelSize(GLenum n_internal_format);

/**
 *	bool Get_GL_PixelType(int n_pixel_type, int &r_n_fmt,
 *		int &r_n_data_type, int &r_n_internal_fmt)
 *		- converts DigiLib pixel type n_pixel_type to OpenGL pixel type
 *		- returns true on success or false in case it couldn't find equivalent format
 */
bool Get_GL_PixelType(int n_pixel_type, int &r_n_fmt, int &r_n_data_type, int &r_n_internal_fmt);

/**
 *	bool CopyImage_To_Texture2D(const ImageStruct *p_image)
 *		- copies image p_image to currently bound texture in active texture unit
 *		- returns true on success, false on failure (bad pixel type)
 */
bool CopyImage_To_Texture2D(const ImageStruct *p_image);

/**
 *	bool CopyTexture2D_To_Image(const ImageStruct *p_image)
 *		- copies image from currently bound texture in active texture unit to p_image
 *		- returns true on success, false on failure (bad pixel type)
 */
bool CopyTexture2D_To_Image(ImageStruct *p_image);

/**
 *	CGLTexture_2D *p_Create_ImageStruct_Texture(CGLState *p_state, const ImageStruct *p_image)
 *		- returns a new texture with size, format and contents of p_image
 *		- p_state is OpenGL state guard
 *		- returns 0 on failure (bad pixel type, not enough memory)
 */
CGLTexture_2D *p_Create_ImageStruct_Texture(CGLState *p_state, const ImageStruct *p_image);

#endif //__IMAGE_TRANSFER_INCLUDED
