#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h> // glFlush(), glFinish()
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include "../GPU_Lib.h"
#include "../../berLame_src/Timer.h"
#include "../../berLame_src/Jpeg.h"
#include "../../berLame_src/Tga.h"
#include "Utils.h"

#if defined(_MSC_VER) && !defined(__MWERKS__)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

int main()
{
	printf("\t\t\t---- GPU-lib speed test ----\n\n");
	printf("This will measure memory speed making some transfers (4 GB of data\n"
		   "to have mean results), then run series of operations, measuring\n"
		   "their run-time and fillrate. All tests shouldn't take more than\n"
		   "couple minutes. Output you see on the screen is simultaneously written\n"
		   "into text file stats.txt. Press enter anytime you feel ready ...\n");
	getchar();

	if(!Init_OpenGL()) {
		fprintf(stderr, "error: OpenGL initialization failed\n");
		return -1;
	}
	// init openGL

	ImageStruct *p_src1, *p_src2, *p_src3, *p_src4, *p_src5, *p_src6, *p_src7;
	if(!(p_src1 = p_LoadImage("_lena_256.jpg")) ||
	   !(p_src2 = p_LoadImage("_lucy_256.jpg")) ||
	   !(p_src3 = p_LoadImage("_grate_a_256.tga")) ||
	   !(p_src4 = p_LoadImage("_arrow.tga")) ||
	   !(p_src5 = p_LoadImage("_edgefinder_a.tga")) ||
	   !(p_src6 = p_CreateGauss(16, 16)) ||
	   !(p_src7 = p_CreateGauss(16, 1))) {
		fprintf(stderr, "error: failed to load source image(s)\n");
		return -1;
	}
	// load src images

	bool b_errors = false;

	ImageStruct *p_dest;
	if(!(p_dest = p_Alloc_Image(256, 256, ImageRGBLinear))) { // use common image
		fprintf(stderr, "error: failed to create mapped image, cannot continue\n");
		return -1;
	}
	// create dest image

	if(b_OpenGLError()) {
		fprintf(stderr, "error: OpenGL errors occured, immediately after init\n");
		b_errors = true;
	}

	FILE *p_fw;
	if(!(p_fw = fopen("stats.txt", "w"))) {
		fprintf(stderr, "warning: failed to open stats.txt for writing\n");
		b_errors = true;
	}

	printf("now measuring CPU / GPU memory troughput\n");
	if(p_fw)
		fprintf(p_fw, "now measuring CPU / GPU memory troughput\n");
	ImageStruct *p_speed_test = p_Alloc_Image(1024, 1024, ImageRGBLinear);
	for(int i = 0; i < 4; i ++) {
		CTimer my_timer;

		double f_start = my_timer.f_Time(), f_saved_time = 0;

		if(i == 0) {
			for(int n = 0; n < 256; ++ n) {
				if(!Upload_ImageStruct(p_speed_test))
					fprintf(stderr, "Upload_ImageStruct()\n");
				double f_start_transfer = my_timer.f_Time();
				glFlush();
				glFinish();
				f_saved_time += my_timer.f_Time() - f_start_transfer;
			}
		} else if(i == 1) {
			for(int n = 0; n < 256; ++ n) {
				if(!Download_ImageStruct(p_speed_test))
					fprintf(stderr, "Download_ImageStruct()\n");
				double f_start_transfer = my_timer.f_Time();
				glFlush();
				glFinish();
				f_saved_time += my_timer.f_Time() - f_start_transfer;
			}
		} else if(i == 2) {
			for(int n = 0; n < 256; ++ n) {
				unsigned int n_fence;
				if(!Upload_Async_ImageStruct(p_speed_test, &n_fence))
					fprintf(stderr, "Upload_Async_ImageStruct()\n");
				double f_start_transfer = my_timer.f_Time();
				if(!WaitFence(n_fence))
					fprintf(stderr, "WaitFence()\n");
				f_saved_time += my_timer.f_Time() - f_start_transfer;
			}
		} else if(i == 3) {
			for(int n = 0; n < 256; ++ n) {
				unsigned int n_fence;
				if(!Download_Async_ImageStruct(p_speed_test, &n_fence))
					fprintf(stderr, "Upload_Async_ImageStruct()\n");
				double f_start_transfer = my_timer.f_Time();
				if(!WaitFence(n_fence))
					fprintf(stderr, "WaitFence()\n");
				f_saved_time += my_timer.f_Time() - f_start_transfer;
			}
		}

		double f_time = my_timer.f_Time();

		double f_speed = 1024.0 * 1024 * 1024 * 4 / (f_time - f_start);
		const char *p_s_units[] = {"", "k", "M", "G"};
		int n_speed_unit = 0;
		while(f_speed > 1000 && n_speed_unit + 1 <
		   sizeof(p_s_units) / sizeof(p_s_units[0])) {
			f_speed *= .001;
			++ n_speed_unit;
		}

		const char *p_test_name_list[] = {
			"synchronous upload",
			"synchronous download",
			"asynchronous upload",
			"asynchronous download",
		};

		if(i > 1) {
			printf("%s: %f %sB/sec, saved time: %f\n",
				p_test_name_list[i], f_speed, p_s_units[n_speed_unit], f_saved_time);
			if(p_fw) {
				fprintf(p_fw, "%s: %f %sB/sec, saved time: %f\n",
					p_test_name_list[i], f_speed, p_s_units[n_speed_unit], f_saved_time);
			}
		} else {
			printf("%s: %f %sB/sec\n",
				p_test_name_list[i], f_speed, p_s_units[n_speed_unit]);
			if(p_fw) {
				fprintf(p_fw, "%s: %f %sB/sec\n",
					p_test_name_list[i], f_speed, p_s_units[n_speed_unit]);
			}
		}

		if(b_OpenGLError()) {
			fprintf(stderr, "error: OpenGL errors occured\n");
			b_errors = true;
		}
	}
	printf("\n");
	if(p_fw)
		fprintf(p_fw, "\n");
	p_speed_test->Optional.Regular.Free(p_speed_test);
	// transfer speed test

	const char *p_test_name_list[] = {
		"ImageImageAdd(p_dest, p_src1, p_src2).bmp",
		"ImageImageMult(p_dest, p_src1, p_src2).bmp",
		"ImageImageSub(p_dest, p_src1, p_src2).bmp",
		"ImageImageDiv(p_dest, p_src1, p_src2).bmp",
		"ImageImageBlend(p_dest, p_src1, p_src2).bmp",
		"ImageImageDecal(p_dest, p_src1, p_src3).bmp",
		"ImageInvert(p_dest, p_src1).bmp",
		"ImageAbs(p_dest, p_src1).bmp",
		"ImageImageMin(p_dest, p_src1, p_src2).bmp",
		"ImageImageMax(p_dest, p_src1, p_src2).bmp",
		"ImageThresh(p_dest, p_src1, .5f).bmp",
		"ImageLog(p_dest, p_src1).bmp",
		"ImageExp(p_dest, p_src1).bmp",
		"ImageSqrt(p_dest, p_src1).bmp",
		"ImagePow(p_dest, p_src1, 2.0f).bmp",
		"ImageScale(p_dest, p_src2, 1.5f, -.5f).bmp",
		"ImageColorMatrix(p_dest, p_src2, p_mat4x4, false).bmp",
		"ImageGreyscale(p_dest, p_src2).bmp",
		"ImageResize(p_dest, p_src2).bmp",
		"ImageRotate(p_dest, p_src2, 30deg, false).bmp",
		"ImageTransform(p_dest, p_src2, ...).bmp",
		"ImageSquareErode(p_dest, p_src1, 8, 8).bmp",
		"ImageSquareDilate(p_dest, p_src1, 8, 8).bmp",
		"ImageSquareOpen(p_dest, p_src1, 3, 3).bmp",
		"ImageSquareClose(p_dest, p_src1, 3, 3).bmp",
		"ImageDiamondErode(p_dest, p_src1, 8, 8).bmp",
		"ImageDiamondDilate(p_dest, p_src1, 8, 8).bmp",
		"ImageDiamondOpen(p_dest, p_src1, 3, 3).bmp",
		"ImageDiamondClose(p_dest, p_src1, 3, 3).bmp",
		"ImageImageErode(p_dest, p_src1, p_src4).bmp",
		"ImageImageDilate(p_dest, p_src1, p_src4).bmp",
		"ImageImageOpen(p_dest, p_src1, p_src4).bmp",
		"ImageImageClose(p_dest, p_src1, p_src4).bmp",
		"ImageImageHitMiss(p_dest, p_src1, p_src5).bmp",
		"ImageMean(p_dest, p_src2, 256, 256).bmp",
		"ImageMedian(p_dest, p_src1, 15).bmp",
		"ImageLocalMin(p_dest, p_src2, 8, 8).bmp",
		"ImageLocalMax(p_dest, p_src2, 8, 8).bmp",
		"ImageConvolveSeparable(p_dest, p_src2, G16, G16).bmp",
		"ImageConvolve2D(p_dest, p_src2, Gauss16).bmp",
		"ImageGetRGBAHistogram(p_src1, 256, ...)",
		"ImageGetMinMax(p_src1, p_minmax)"
	};

	printf("                    operation                  "
		"time to process 256 x 256 8bit RGBA      fillrate\n");
	if(p_fw) {
		fprintf(p_fw, "                    operation                  "
			"time to process 256 x 256 8bit RGBA      fillrate\n");
	}
	// print table header

	Disable_AutoUploads();
	Disable_AutoDownloads();
	// don't want to add transfer times to processing time

	CTimer my_timer;
	for(int i = 0; i < sizeof(p_test_name_list) / sizeof(p_test_name_list[0]) && !b_errors; ++ i) {
		char p_s_buffer[256];
		assert(strlen(p_test_name_list[i]) < 256);
		strcpy(p_s_buffer, p_test_name_list[i]);
		if(strstr(p_s_buffer, ".bmp"))
			*(char*)strstr(p_s_buffer, ".bmp") = 0;
		printf("%-60s", p_s_buffer);
		// print test name, erase .bmp suffix

		my_timer.ResetTimer();
		int n_frame_counter = 0;
		double f_last_time;
		while((f_last_time = my_timer.f_Time()) < 3) {
			switch(i) {
			case 0:
				ImageImageAdd(p_dest, p_src1, p_src2);
				break;
			case 1:
				ImageImageMult(p_dest, p_src1, p_src2);
				break;
			case 2:
				ImageImageSub(p_dest, p_src1, p_src2);
				break;
			case 3:
				ImageImageDiv(p_dest, p_src1, p_src2);
				break;
			case 4:
				ImageImageBlend(p_dest, p_src1, p_src2, .5f, .6f);
				break;
			case 5:
				ImageImageDecal(p_dest, p_src1, p_src3);
				break;
			case 6:
				ImageInvert(p_dest, p_src1);
				break;
			case 7:
				ImageAbs(p_dest, p_src1);
				break;
			case 8:
				ImageImageMin(p_dest, p_src1, p_src2);
				break;
			case 9:
				ImageImageMax(p_dest, p_src1, p_src2);
				break;
			case 10:
				ImageThresh(p_dest, p_src1, .5f);
				break;
			case 11:
				ImageLog(p_dest, p_src1);
				break;
			case 12:
				ImageExp(p_dest, p_src1);
				break;
			case 13:
				ImageSqrt(p_dest, p_src1);
				break;
			case 14:
				ImagePow(p_dest, p_src1, 2.0f);
				break;
			case 15:
				ImageScale(p_dest, p_src2, 1.5f, -.2f);
				break;
			case 16:
				{
					static const float p_mat4x4[] = {
						.299f, .587f, .114f, .956f * .18f + .621f * -.001f,
						.299f, .587f, .114f, -.272f * .18f - .647f * -.001f,
						.299f, .587f, .114f, -1.105f * .18f + 1.702f * -.001f,
						0, 0, 0, 1
					}; // simple sepia filter
					ImageColorMatrix(p_dest, p_src2, p_mat4x4, false);
				}
				break;
			case 17:
				ImageGreyscale(p_dest, p_src2);
				break;
			case 18:
				ImageResize(p_dest, p_src2);
				break;
			case 19:
				{
					const float f_pi = 3.141592653589794626433832795028841971697510f;
					ImageRotate(p_dest, p_src2, 5 * f_pi / 180, false, 1);
				}
				break;
			case 20:
				ImageTransform(p_dest, p_src2, -.8f, 1, -1, -1, 1, -1, .8f, 1);
				break;
			case 21:
				ImageSquareErode(p_dest, p_src1, 8, 8);
				break;
			case 22:
				ImageSquareDilate(p_dest, p_src1, 8, 8);
				break;
			case 23:
				ImageSquareOpen(p_dest, p_src1, 3, 3);
				break;
			case 24:
				ImageSquareClose(p_dest, p_src1, 3, 3);
				break;
			case 25:
				ImageDiamondErode(p_dest, p_src1, 8, 8);
				break;
			case 26:
				ImageDiamondDilate(p_dest, p_src1, 8, 8);
				break;
			case 27:
				ImageDiamondOpen(p_dest, p_src1, 3, 3);
				break;
			case 28:
				ImageDiamondClose(p_dest, p_src1, 3, 3);
				break;
			case 29:
				ImageImageErode(p_dest, p_src1, p_src4);
				break;
			case 30:
				ImageImageDilate(p_dest, p_src1, p_src4);
				break;
			case 31:
				ImageImageOpen(p_dest, p_src1, p_src4);
				break;
			case 32:
				ImageImageClose(p_dest, p_src1, p_src4);
				break;
			case 33:
				ImageImageHitMiss(p_dest, p_src1, p_src5);
				break;
			case 34:
				ImageMean(p_dest, p_src2, 256, 256);
				break;
			case 35:
				ImageMedian(p_dest, p_src1, 15);
				break;
			case 36:
				ImageLocalMin(p_dest, p_src1, 8, 8);
				break;
			case 37:
				ImageLocalMax(p_dest, p_src1, 8, 8);
				break;
			case 38:
				ImageConvolveSeparable(p_dest, p_src2, p_src7, p_src7);
				// must use another, normalized gaussian kernel;
				// could use int16 or like that to speed-up
				break;
			case 39:
				ImageConvolve2D(p_dest, p_src2, p_src6);
				break;
			case 40:
				{
					unsigned int p_histogram_r[256];
					unsigned int p_histogram_g[256];
					unsigned int p_histogram_b[256];
					unsigned int p_histogram_a[256];
					ImageGetRGBAHistogram(p_src1, 256,
						p_histogram_r, p_histogram_g, p_histogram_b, p_histogram_a);
				}
				break;
			case 41:
				{
					unsigned int p_minmax[2];
					ImageGetMinMax(p_src1, p_minmax);
				}
				break;
			}
			if(!n_frame_counter)
				my_timer.ResetTimer(); // don't count first run
			++ n_frame_counter;
		}

		glFlush();
		glFinish();

		f_last_time = my_timer.f_Time();

		double f_frame_time = f_last_time / n_frame_counter;
		double f_fillrate = (256 * 256) / f_frame_time;
		const char *p_s_units[] = {"", "k", "M", "G"};
		int n_fillrate_unit = 0;
		while(f_fillrate > 1000 && n_fillrate_unit + 1 <
		   sizeof(p_s_units) / sizeof(p_s_units[0])) {
			f_fillrate *= .001;
			++ n_fillrate_unit;
		}
		if(p_fw) {
			fprintf(p_fw, "%-60s %10f ms, %20f %spix/s\n", p_test_name_list[i],
				f_frame_time * 1000, f_fillrate, p_s_units[n_fillrate_unit]);
		}
		printf("%10f ms, %20f %spix/s\n", f_frame_time * 1000,
			f_fillrate, p_s_units[n_fillrate_unit]);
		// print benchmark information into the logfile

		if(b_AsyncTransfersSupported()) {
			unsigned int n_fence_id;
			if(!Download_Async_ImageStruct(p_dest, &n_fence_id)) {
				fprintf(stderr, "error: Download_Async_ImageStruct failed\n");
				b_errors = true;
			}
			if(b_NVFenceSupported())
				WaitFence(n_fence_id);
		} else {
			if(!Download_ImageStruct(p_dest)) {
				fprintf(stderr, "error: Download_ImageStruct failed\n");
				b_errors = true;
			}
		}
		// transfers are disabled, we need to get our result manually

		if(i < 40 && !Save_RGBLinearImageStruct(p_test_name_list[i], p_dest)) {
			fprintf(stderr, "error : error writing image \'%s\'\n", p_test_name_list[i]);
			b_errors = true;
		}
		// save result preview

		if(b_OpenGLError()) {
			fprintf(stderr, "error: OpenGL errors occured\n");
			b_errors = true;
		}
	}
	// operation speed test

	Enable_AutoDownloads();
	// enable downloads to make the following code simpler

	for(int n_pass = 0; n_pass < 2; ++ n_pass) {
		printf((n_pass)? "equalized histogram:\n" :
			"\nand now comes histogram equalization test (histogram of lena_256.jpg):\n");
		PrintHistogram(stdout, (n_pass)? p_dest : p_src1);
		printf("\n");
		if(p_fw) {
			fprintf(p_fw, (n_pass)? "equalized histogram:\n" :
				"\nand now comes histogram equalization test (histogram of lena_256.jpg):\n");
			PrintHistogram(p_fw, (n_pass)? p_dest : p_src1);
			fprintf(p_fw, "\n");
		}
		// print image histogram (feature test)

		if(b_OpenGLError()) {
			fprintf(stderr, "error: OpenGL histogram errors occured\n");
			b_errors = true;
		}

		unsigned long p_minmax[4];
		ImageGetMinMax((n_pass)? p_dest : p_src1, p_minmax);
		printf("ImageGetMinMax(%s) = ", (n_pass)? "p_dest" : "p_src1");
		PrintARGB(stdout, p_minmax[0]);
		printf(", ");
		PrintARGB(stdout, p_minmax[1]);
		printf("\n\n");
		if(p_fw) {
			fprintf(p_fw, "ImageGetMinMax(%s) = ", (n_pass)? "p_dest" : "p_src1");
			PrintARGB(p_fw, p_minmax[0]);
			fprintf(p_fw, ", ");
			PrintARGB(p_fw, p_minmax[1]);
			fprintf(p_fw, "\n\n");
		}
		// print image min and max value (feature test)

		if(b_OpenGLError()) {
			fprintf(stderr, "error: OpenGL minmax errors occured\n");
			b_errors = true;
		}

		if(n_pass)
			break;

		float f_min_r = ((p_minmax[0] >> 16) & 0xff) / 255.0f;
		float f_min_g = ((p_minmax[0] >> 8) & 0xff) / 255.0f;
		float f_min_b = (p_minmax[0] & 0xff) / 255.0f;
		float f_max_r = ((p_minmax[1] >> 16) & 0xff) / 255.0f;
		float f_max_g = ((p_minmax[1] >> 8) & 0xff) / 255.0f;
		float f_max_b = (p_minmax[1] & 0xff) / 255.0f;
		float p_equalize_matrix[] = {
			1.0f / (f_max_r - f_min_r), 0, 0, -f_min_r / (f_max_r - f_min_r),
			0, 1.0f / (f_max_g - f_min_g), 0, -f_min_g / (f_max_g - f_min_g),
			0, 0, 1.0f / (f_max_b - f_min_b), -f_min_b / (f_max_b - f_min_b),
			0, 0, 0, 1
		};
		// create equalization matrix

		ImageColorMatrix(p_dest, p_src1, p_equalize_matrix, false);
		// multiply image

		if(!Save_RGBLinearImageStruct("lena_histeq.bmp", p_dest)) {
			fprintf(stderr, "error : error writing image \'%s\'\n", "lena_histeq.bmp");
			b_errors = true;
		}
		// save it
	}
	// equalize lena image (feature test)

	printf("have %d imagestructs\n", n_GPU_ImageStruct_Num());
	if(p_fw)
		fprintf(p_fw, "have %d imagestructs\n", n_GPU_ImageStruct_Num());

	p_dest->Optional.Regular.Free(p_dest);
	p_src1->Optional.Regular.Free(p_src1);
	p_src2->Optional.Regular.Free(p_src2);
	p_src3->Optional.Regular.Free(p_src3);
	p_src4->Optional.Regular.Free(p_src4);
	p_src5->Optional.Regular.Free(p_src5);
	p_src6->Optional.Regular.Free(p_src6);
	p_src7->Optional.Regular.Free(p_src7);
	// delete images

	if(b_OpenGLError()) {
		fprintf(stderr, "error: OpenGL errors occured\n");
		b_errors = true;
	}

	printf("shutting down ...\n");
	printf("have %d imagestructs\n", n_GPU_ImageStruct_Num());
	printf("have %d render-buffers\n", n_Shader_Num());
	printf("have %d shaders\n", n_RenderBuffer_Num());
	printf("have %d temporal (%.2f kB) textures\n",
		n_TempTexture_Num(), n_TempTexture_Size() / 1024.0);
	if(p_fw) {
		fprintf(p_fw, "shutting down ...\n");
		fprintf(p_fw, "have %d imagestructs\n", n_GPU_ImageStruct_Num());
		fprintf(p_fw, "have %d render-buffers\n", n_Shader_Num());
		fprintf(p_fw, "have %d shaders\n", n_RenderBuffer_Num());
		fprintf(p_fw, "have %d temporal (%.2f kB) textures\n",
			n_TempTexture_Num(), n_TempTexture_Size() / 1024.0);
		if(b_errors)
			fprintf(p_fw, "\nerror(s) occured\n");
	}

	if(p_fw)
		fclose(p_fw);

	if(!Shutdown_OpenGL()) {
		fprintf(stderr, "error: OpenGL shutdown failed");
		return -1;
	}

	return 0;
}

/*
 *		-end-of-file-
 */
