/*
								+----------------------------------+
								|                                  |
								|   *** Temp texture manager ***   |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|         TempTexManager.h         |
								|                                  |
								+----------------------------------+
*/

#ifndef __GL_TEXTURE_2D_MANAGER_INCLUDED
#define __GL_TEXTURE_2D_MANAGER_INCLUDED

#include "Manager.h"

/**
 *	struct TTextureFormat
 *		- simple texture format structure
 *		- contains texture resolution and pixel format
 */
struct TTextureFormat {
	int n_width, n_height;
	int n_internal_format;

	inline TTextureFormat()
	{}

	inline TTextureFormat(int _n_width, int _n_height, int _n_internal_format)
		:n_width(_n_width), n_height(_n_height), n_internal_format(_n_internal_format)
	{}

	inline TTextureFormat(const CGLTexture_2D *p_texture)
		:n_width(p_texture->n_Width()), n_height(p_texture->n_Height()),
		n_internal_format(p_texture->n_Pixel_Format())
	{}
};

/**
 *	class CGLNamedTexture_2D
 *		- extension of render-buffer with automatically assigned id
 */
class CGLNamedTexture_2D : public CGLTexture_2D {
protected:
	static int m_n_id_space;
	int m_n_id;

public:
	inline int n_Id() const { return m_n_id; }

	inline CGLNamedTexture_2D(CGLState *p_state, int n_width, int n_height, GLenum n_internal_format,
		bool b_create_mipmaps = true, int n_border_width = 0,
		GLenum n_format = GL_RGB, GLenum n_data_type = GL_UNSIGNED_BYTE,
		const void *p_src_data = 0)
		:CGLTexture_2D(p_state, n_width, n_height, n_internal_format,
		b_create_mipmaps, n_border_width, n_format, n_data_type, p_src_data)
	{
		m_n_id = ++ m_n_id_space;
	}
};

//int CGLNamedTexture_2D::m_n_id_space = 0;
// needs to be declared once

extern unsigned int n_GL_PixelSize(GLenum n_internal_format);

/**
 *	class CGetTextureCost
 *		- cost evaluator for render-buffer manager purging purposes
 *		- currently returns size of buffer surface in pixels
 */
class CGetTextureCost {
public:
	inline int n_Cost(const CGLNamedTexture_2D *p_tex) const
	{
		return p_tex->n_Width() * p_tex->n_Height() *
			n_GL_PixelSize(p_tex->n_Pixel_Format()); // todo - calculate with pixel size
	}
};

/**
 *	class CFindTex_Format
 *		- finds texture format, equal to another texture
 */
class CFindTex_Format {
protected:
	const TTextureFormat m_t_format;

public:
	inline CFindTex_Format(const TTextureFormat &r_format)
		:m_t_format(r_format)
	{}

	inline bool operator ()(const TTextureFormat &r_format) const
	{
		return r_format.n_width == m_t_format.n_width &&
			r_format.n_height == m_t_format.n_height &&
			r_format.n_internal_format == m_t_format.n_internal_format;
	}
};

/**
 *	class CTextureFactory
 *		- texture factory
 */
class CTextureFactory {
protected:
	CGLState *m_p_state;

public:
	inline CTextureFactory(CGLState *p_state)
		:m_p_state(p_state)
	{}

	inline CGLNamedTexture_2D *Load(const TTextureFormat &r_format)
	{
		CGLNamedTexture_2D *p_texture;
		if(!(p_texture = new CGLNamedTexture_2D(m_p_state, r_format.n_width,
		   r_format.n_height, r_format.n_internal_format, false)))
			return 0;

		p_texture->r_Parameters(m_p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
		p_texture->r_Parameters(m_p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
		// clamp to edge

		return p_texture;
	}

	static inline bool b_Valid_Object(const CGLNamedTexture_2D *p_object)
	{
		return p_object != 0;
	}
};

/**
 *	class CGLTexture_2D_Manager
 *		- OpenGL 2d-texture manager class
 */
typedef CResourceManager<CGLNamedTexture_2D, CGetTextureCost, int, CFindIntId<CGLNamedTexture_2D>,
	TTextureFormat, CFindTex_Format, CTextureFactory> CGLTexture_2D_Manager;

#endif //__GL_TEXTURE_2D_MANAGER_INCLUDED
